/***************************************************************************
 *   Copyright (C) 2008 by Sergey Shilov                                   *
 *   shilov_serg@users.sourceforge.net                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/time.h>
//#include <asm/ioctls.h>
#include <linux/serial.h>
#include <ctype.h>
#include <stdio.h>
#ifdef	C_LINUX
#include <stdio_ext.h>
#endif
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <errno.h>
#include <syslog.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>
#include <getopt.h>
#include <termios.h>
#include <pthread.h>
#include <iconv.h>

#define DEF_DEVICE_NAME "/dev/ttyS0"
#define DEF_CONFIG_FILE "/etc/ttytest.conf"
#define DEF_OUTPUT_FILE ""

#define MODE_LOOPBACK 	0
#define MODE_READ 	1
#define MODE_STUB 	2
#define MODE_TERMINAL 	3
#define MODE_WRITE 	4


#define FLAG_NONE		0
#define FLAG_PRINT_OUTPUT	1
#define FLAG_IN_MAIN_THREAD	1 << 4

#define DEF_PAUSE 0
#define DEF_SEND_INTERVAL 0
#define DEF_LOOPBACK 0
#define DEF_TTY_BAUD B9600

// 1..8
#define DATABITS_1_MASK 0x01
#define DATABITS_2_MASK 0x03
#define DATABITS_3_MASK 0x07
#define DATABITS_4_MASK 0x0F
#define DATABITS_5_MASK 0x1F
#define DATABITS_6_MASK 0x3F
#define DATABITS_7_MASK 0x7F
#define DATABITS_8_MASK 0xFF

#define DEF_TTY_DATABITS 8

// 1, 1.5, 2
#define DEF_TTY_STOPB 1
// 00 = NONE, 01 = Odd, 02 = Even, 03 = Mark, 04 = Space
#define DEF_TTY_PARITY 0
#define DEF_TTY_CRTSCTS 0
#define DEF_TTY_XONXOFF 0
#define DEF_TTY_DSR 0

#ifdef	C_LINUX
# include <libintl.h>
# include <locale.h>
# define	gettext_noop(s)	s
# define	_(s)	gettext(s)
# define	N_(s)	gettext_noop(s)
#else
# define	_(s)	s
# define	N_(s)	s
#endif

#define THREAD_RX  0
#define THREAD_TX  1

//
// Program options
//
int mode = MODE_LOOPBACK;

const char * DEF_DATA_STRING = "test test test test test test";

static char * devicename = NULL ;
static char * config_file = NULL;
static char * output_file = NULL;
static char * data_str = NULL;
static char * dev_codepage = NULL;
static char * tty_codepage = NULL;
int verbose = 0;
int no_offset = 0;
int start_pause = DEF_PAUSE;
int enum_strings = 0;
int hex_print = 0;
float send_interval = DEF_SEND_INTERVAL;
int queue_size = -1;
int adaptive_tx = 0;
int wait_tx = 0;
int show_nonprint = 0;
int use_mtk_reg_prefix = 0;

//
// Serial device parameters
//
long baud_rate = 9600;            // baud rate DECIMAL value from command line
long BAUD = DEF_TTY_BAUD;         // normalized to termios.h consts BAUD value
long DATABITS = DEF_TTY_DATABITS;
float VSTOPBITS = 0;
long STOPBITS = DEF_TTY_STOPB;
long PARITYON = 0;
long PARITY = DEF_TTY_PARITY;
int use_RTSCTS = 0;
int use_XonXoff = 0;
int use_DSR = 0;

int device_fd = -1;
int output_fd = -1;
FILE *device_r = NULL;
FILE *device_w = NULL;
FILE *input = NULL;
FILE *output = NULL;


//
// Thread specific variables
//
struct thread_args
{
	unsigned int flags;
	unsigned int mode;
	FILE *in;
	FILE *out;
	FILE *dev_r;
	FILE *dev_w;
};
// Attributes
pthread_attr_t tx_tattr;
pthread_attr_t rx_tattr;
// IDs
pthread_t tx_tid = ( pthread_t ) (-1);
pthread_t rx_tid = ( pthread_t ) (-1);
// Shared FDs mutexes
pthread_mutex_t STOP_RX_mutex;
pthread_mutex_t STOP_TX_mutex;
pthread_mutex_t output_mutex;

//
// Thread control variables
//
volatile int STOP=0;
volatile int STOP_RX=0;
volatile int STOP_TX=0;




#define DATABITS_0_MASK 0x00
#define DATABITS_1_MASK 0x01
#define DATABITS_2_MASK 0x03
#define DATABITS_3_MASK 0x07
#define DATABITS_4_MASK 0x0F
#define DATABITS_5_MASK 0x1F
#define DATABITS_6_MASK 0x3F
#define DATABITS_7_MASK 0x7F
#define DATABITS_8_MASK 0xFF


// Streams conversion variables
//
// Handlers of IO characters convertors
//
iconv_t conv_tty2dev = ( iconv_t )( -1 );
iconv_t conv_dev2tty = ( iconv_t )( -1 );
//
// Token of "MTK2" charset
int MTK2_IO = 0;


//-------------------------------------------
#define MTK_RUS  0x00        //REG RUS
#define MTK_CIF  0x1B        //REG DIGIT
#define MTK_LAT  0x1F        //REG LAT
#define MTK_PB   0x04        //' '
#define MTK_MALT 0x09        //KTM
#define MTK_RAW  0x1E        //=
#define MTK_DW   0x0E        //:
#define MTK_0    0x16        //
#define MTK_1    0x17        //1
#define MTK_2    0x13        //2
#define MTK_3    0x01        //3
#define MTK_5    0x10        //5
#define MTK_6    0x15        //6
#define MTK_7    0x07        //7
#define MTK_8    0x06        //8
#define MTK_9    0x18        //9
#define MTK_CR   0x08        // \r
#define MTK_LF   0x02        // \n

#define MTK_A    0x03
#define MTK_B    0x19
#define MTK_W    0x13
#define MTK_G    0x1A
#define MTK_D    0x09
#define MTK_E    0x01
#define MTK_GE   0x1E
#define MTK_Z    0x11
#define MTK_I    0x06
#define MTK_J    0x0B
#define MTK_K    0x0F
#define MTK_L    0x12
#define MTK_M    0x1C
#define MTK_N    0x0C
#define MTK_O    0x18
#define MTK_P    0x16
#define MTK_R    0x0A
#define MTK_S    0x05
#define MTK_T    0x10
#define MTK_U    0x07
#define MTK_F    0x0D
#define MTK_HA   0x14
#define MTK_C    0x0E
#define MTK_4    0x0A
#define MTK_SH   0x1A
#define MTK_SHA  0x14
#define MTK_TZ   0x05
#define MTK_Y    0x15
#define MTK_MZ   0x1D
#define MTK_EE   0x0D
#define MTK_JU   0x0B
#define MTK_JA   0x0B


#define REG_MTK_UNDEFINED  0xFF

#define REG_MTK_MASK  0xC0

#define MTK_BYTE_MASK 0x1F

#define BIT_MTK_CIF  0x00
#define BIT_MTK_RUS  0x40
#define BIT_MTK_LAT  0x80

#define IS_MTK_REG( X ) \
(\
	( X == MTK_RUS) || \
	( X == MTK_LAT) || \
	( X == MTK_CIF ) \
)

char CP866_MTK2[256] = {
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
                           0x1c, 0xc4, 0xc2, 0x1c, 0x1c, 0xc8, 0x1c, 0x1c,
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
                           0xc4, 0x1c, 0x05, 0x09, 0x1c, 0x1c, 0x1c, 0x05,
                           0x0f, 0x12, 0x09, 0x11, 0x0c, 0x03, 0x1c, 0x1d,
                           0x16, 0x17, 0x13, 0x01, 0x0a, 0x10, 0x15, 0x07,
                           0x06, 0x18, 0x0e, 0x1c, 0x1c, 0x1e, 0x1c, 0x19,
                           0x1c, 0x83, 0x99, 0x8e, 0x89, 0x81, 0x8d, 0x9a,
                           0x94, 0x86, 0x8b, 0x8f, 0x92, 0x9c, 0x8c, 0x98,
                           0x96, 0x97, 0x8a, 0x85, 0x90, 0x87, 0x9e, 0x93,
                           0x9d, 0x95, 0x91, 0x1c, 0x1c, 0x1c, 0x1c, 0x03,
                           0x05, 0x83, 0x99, 0x8e, 0x89, 0x81, 0x8d, 0x9a,
                           0x94, 0x86, 0x8b, 0x8f, 0x92, 0x9c, 0x8c, 0x98,
                           0x96, 0x97, 0x8a, 0x85, 0x90, 0x87, 0x9e, 0x93,
                           0x9d, 0x95, 0x91, 0x1c, 0x0E, 0x1c, 0x1c, 0x1c,
                           0x43, 0x59, 0x53, 0x5a, 0x49, 0x41, 0x5e, 0x51,
                           0x46, 0x4b, 0x4f, 0x52, 0x5c, 0x4c, 0x58, 0x56,
                           0x4a, 0x45, 0x50, 0x47, 0x4d, 0x54, 0x4e, 0x0a,
                           0x1a, 0x14, 0x05, 0x55, 0x5d, 0x0d, 0x0b, 0x57,
                           0x43, 0x59, 0x53, 0x5a, 0x49, 0x41, 0x5e, 0x51,
                           0x46, 0x4b, 0x4f, 0x52, 0x5c, 0x4c, 0x58, 0x56,
                           0x1c, 0x1c, 0x1c, 0x0E, 0x1c, 0x1c, 0x1c, 0x1c,
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
                           0x4a, 0x45, 0x50, 0x47, 0x4d, 0x54, 0x4e, 0x0a,
                           0x1a, 0x14, 0x05, 0x55, 0x5d, 0x0d, 0x0b, 0x57,
                           0x1c, 0x1c, 0x81, 0x81, 0x86, 0x86, 0x1c, 0x1c,
                           0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c
                       };

unsigned char MTK2_CP866[3][32] = {
                                      { //MTK_CIF
                                          0xff,  0x33,   0x0a,  0x2d,   0x20,   0x60,   0x38,  0x37,
                                          0x0d,  0x2a,   0x97,  0x9e,   0x2c,   0x9d,   0x3a,  0x28,
                                          0x35,  0x2b,   0x29,  0x32,   0x99,   0x36,   0x30,  0x31,
                                          0x39,  0x3f,   0x98,  0xff,   0x2e,   0x2f,   0x3d,  0xff
                                      },
                                      { //MTK_RUS
                                          0xff,  0x85,   0x0a,  0x80,   0x20,   0x91,   0x88,  0x93,
                                          0x0d,  0x84,   0x90,  0x89,   0x8d,   0x94,   0x96,  0x8a,
                                          0x92,  0x87,   0x8b,  0x82,   0x95,   0x9b,   0x8f,  0x9f,
                                          0x8e,  0x81,   0x83,  0xff,   0x8c,   0x9c,   0x86,  0xff
                                      },
                                      { //MTK_LAT
                                          0xff,  0x27,   0x0a,  0x41,   0x20,   0x53,   0x49,  0x55,
                                          0x0d,  0x44,   0x52,  0x4a,   0x4e,   0x46,   0x43,  0x4b,
                                          0x54,  0x5a,   0x4c,  0x57,   0x48,   0x59,   0x50,  0x51,
                                          0x4f,  0x42,   0x47,  0xff,   0x4d,   0x58,   0x56,  0xff
                                      }
                                  };

const unsigned char MTK_REGS[3] = { MTK_CIF, MTK_RUS, MTK_LAT};

struct mtk_reg_marker
{
	unsigned char reg_prefix;
	unsigned char prefix_char;
	char reg_chars[4];
};


struct cmtk_conv_data
{
	unsigned char mtk_reg;
	unsigned char mtk_reg_index;
	unsigned int flag;
	unsigned int UkrRegCount;
	unsigned int Reg4Count;
	unsigned int cntCrLf;
	char BufCrLf[8];
	struct mtk_reg_marker regs[3];
};

const struct cmtk_conv_data def_cmtk_conv_data =
    {
	    0,0,0,0,0,0,
	    {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
	    {
	        { 0, '^', {'%','C',0x00 ,0x00}}, // CIF
	        { 0, '^', {'!','R',0x00 ,0x00}}, // RUS
	        { 0, '^', {'*','L',0x00 ,0x00}}, // LAT
	    }
    };


typedef struct cmtk_conv_data cmtk_conv_data_t;
typedef cmtk_conv_data_t * p_cmtk_conv_data;

int cmtk_strlen ( unsigned char *mtk_str )
{
	int i;
	for( i = 0; ( ( mtk_str[i] != 0xFF ) && ( i < 074000 ) ); i++ )
		;
	return i;
}

unsigned char cmtk_byte_CP866toMTK2( unsigned char *mtk_reg, unsigned char c )
{
	unsigned char res;

	res = CP866_MTK2[ c ];
	switch( res & REG_MTK_MASK )
	{
		case BIT_MTK_CIF   :
			if( *mtk_reg == MTK_CIF )
				break;
			*mtk_reg = MTK_CIF;
			return MTK_CIF;

		case BIT_MTK_RUS:
			if( *mtk_reg == MTK_RUS )
				break;
			*mtk_reg = MTK_RUS;
			return MTK_RUS;

		case BIT_MTK_LAT:
			if( *mtk_reg == MTK_LAT )
				break;
			*mtk_reg = MTK_LAT;
			return MTK_LAT;
	}
	return res & MTK_BYTE_MASK;
}

char cmtk_byte_MTK2toCP866( unsigned char* mtk_reg, unsigned char m )
{
	switch( m )
	{
		case MTK_CIF:
			*mtk_reg = MTK_CIF;
			return 0;

		case MTK_RUS:
			*mtk_reg = MTK_RUS;
			return 0;

		case MTK_LAT:
			*mtk_reg = MTK_LAT;
			return 0;
		default:
			break;
	}
	switch( *mtk_reg )
	{
		case MTK_CIF:
			return MTK2_CP866[0][m];

		case MTK_RUS:
			return MTK2_CP866[1][m];

		case MTK_LAT:
			return MTK2_CP866[2][m];
	}
	return 0;
}


long int cmtk_buf_CP866toMTK2( p_cmtk_conv_data conv_data,
                               unsigned char *bufCP866,
                               long int bufCP866_len,
                               unsigned char *bufMTK2,
                               long int bufMTK2_len)
{
	unsigned char m;
	long int i, j;
	char  cur = 0x00;

	for( j = 0, i = 0; ( ( j < bufCP866_len ) && ( i < bufMTK2_len ) ); j++ )
	{
		if( cur == 0x0D )
		{
			conv_data->flag = 1;
		}
		if( conv_data->flag == 1 )
		{
			if ( bufCP866[j] == 0x0D )
				continue;
			else if ( bufCP866[j] != 0x0A )
				conv_data->flag = 0;
		}
		cur = bufCP866[j];
		int r = 0, next_char = 0;
		for (  r = 0; r < 3 ; r++ )
		{
			if ( ( conv_data->regs[r].prefix_char != 0x00 ) && ( cur == conv_data->regs[r].prefix_char ) )
			{
				conv_data->regs[r].reg_prefix = 1;
				next_char++;
				break;
			}
			if ( ( conv_data->regs[r].prefix_char == 0x00 ) || ( conv_data->regs[r].reg_prefix  == 1 ) )
			{
				conv_data->regs[r].reg_prefix = 0;
				if ( strchr( ( char * ) &conv_data->regs[r].reg_chars,  cur ) != NULL )
				{
					bufMTK2[i++] = MTK_REGS[r];
					conv_data->mtk_reg = MTK_REGS[r];
					next_char++;
					break;
				}
			}
		}
		if ( next_char > 0 )
			continue;
		if (bufCP866[j] == '4')
		{
			bufMTK2[i++] = MTK_CIF;
			bufMTK2[i++] = MTK_LAT;
			bufMTK2[i++] = MTK_CIF;
			bufMTK2[i++] = MTK_4;
			conv_data->mtk_reg = MTK_CIF;
			continue;
		}
		m = cmtk_byte_CP866toMTK2( &conv_data->mtk_reg, bufCP866[j] );
		bufMTK2[i++] = m;
		if( IS_MTK_REG( m ) )
			j--;
	}
	return i;
}

long int cmtk_buf_MTK2toCP866( p_cmtk_conv_data conv_data,
                               unsigned char *bufMTK2,
                               long int bufMTK2_len,
                               char *bufCP866,
                               long int bufCP866_len )
{
	cmtk_conv_data_t cdata = def_cmtk_conv_data;
	p_cmtk_conv_data p_cdata = &cdata;

	long int i, j ;

	if ( conv_data != NULL )
		p_cdata = conv_data;

	for( j = 0, i = 0 ; ( ( j < bufMTK2_len ) && ( i < bufCP866_len ) ); j++ )
	{
		switch( bufMTK2[j] )
		{
			case MTK_CIF:
				{
					p_cdata->mtk_reg = bufMTK2[j] ;
					if (p_cdata->UkrRegCount == 1)
						p_cdata->UkrRegCount++;
					else
						p_cdata->UkrRegCount  = 0;
					if (p_cdata->Reg4Count == 0 || p_cdata->Reg4Count == 2)
						p_cdata->Reg4Count++;
					else
						p_cdata->Reg4Count = 0;
					//if( p_cdata->mtk_reg_index  == 0 )
					//	continue;
					p_cdata->mtk_reg_index  = 0;
					if ( p_cdata->flag == 1 )
					{
						bufCP866[i++] = '^';
						bufCP866[i++] = 'C';
					}
					continue;
				}
			case MTK_RUS:
				{
					p_cdata->mtk_reg = bufMTK2[j] ;
					p_cdata->Reg4Count = 0;
					if (p_cdata->UkrRegCount == 0)
						p_cdata->UkrRegCount++;
					else
						p_cdata->UkrRegCount = 0;
					//if ( p_cdata->mtk_reg_index  == 1 )
					//	continue;
					p_cdata->mtk_reg_index  = 1;
					if ( p_cdata->flag == 1 )
					{
						bufCP866[i++] = '^';
						bufCP866[i++] = 'R';
					}
					continue;
				}
			case MTK_LAT:
				{
					p_cdata->mtk_reg = bufMTK2[j] ;
					if ( p_cdata->Reg4Count == 1 )
						p_cdata->Reg4Count++;
					else
						p_cdata->Reg4Count = 0;
					if ( p_cdata->UkrRegCount == 2 )
						p_cdata->UkrRegCount++;
					else
						p_cdata->UkrRegCount = 0;
					//if( p_cdata->mtk_reg_index == 2 )
					//	continue;
					p_cdata->mtk_reg_index = 2;
					if ( p_cdata->flag == 1 )
					{
						bufCP866[i++] = '^';
						bufCP866[i++] = 'L';
					}
					continue;
				}
			case MTK_LF:
				{
					p_cdata->BufCrLf[p_cdata->cntCrLf++] = MTK_LF;
					if (p_cdata->cntCrLf == 2)
						break;
					continue;
				}
			case MTK_CR:
				{
					p_cdata->BufCrLf[p_cdata->cntCrLf++] = MTK_CR;
					if (p_cdata->cntCrLf == 2)
						break;
					continue;
				}
		}
		if (p_cdata->cntCrLf != 0)
		{
			p_cdata->BufCrLf[p_cdata->cntCrLf] = 0x00;
			if (p_cdata->cntCrLf == 1)
			{
				bufCP866[i++] = 0x0D;
				bufCP866[i++] = 0x0A;
				p_cdata->BufCrLf[0] = 0x00;
				p_cdata->cntCrLf = 0;
			}
			else
			{
				if (p_cdata->BufCrLf[0] == p_cdata->BufCrLf[1])
				{
					bufCP866[i++] = 0x0D;
					bufCP866[i++] = 0x0A;
					p_cdata->BufCrLf[0] = 0x00;
					p_cdata->cntCrLf = 0;
					j--;
					continue;
				}
				else
				{
					bufCP866[i++] = 0x0D;
					bufCP866[i++] = 0x0A;
					p_cdata->BufCrLf[0] = 0x00;
					p_cdata->cntCrLf = 0;
				}
			}
		}
		if (p_cdata->UkrRegCount == 3 && bufMTK2[j] == MTK_3)
		{
			bufCP866[i++] = 242; // 0xf2 UKRAINIAN IE
			p_cdata->UkrRegCount = 0;
			continue;
		}
		else if (p_cdata->UkrRegCount == 3 && bufMTK2[j] == MTK_8)
		{
			bufCP866[i++] = 244; // 0xf4 UKRAINIAN ???
			p_cdata->UkrRegCount = 0;
			continue;
		}
		else if (p_cdata->Reg4Count == 3 && bufMTK2[j] == MTK_4)
		{
			bufCP866[i++] = '4';
			p_cdata->Reg4Count = 0;
			continue;
		}
		else
		{
			p_cdata->Reg4Count = 0;
			p_cdata->UkrRegCount = 0;
		}
		if ( ( bufMTK2[j] != MTK_CR ) && ( bufMTK2[j] != MTK_LF ) )
			bufCP866[i++] = MTK2_CP866[p_cdata->mtk_reg_index][bufMTK2[j]];
	}
	return i;
}

static void free_and_NULL ( const void **P )
{
	if ( ( P != NULL ) && ( *P != NULL ) )
	{
		free ( ( void * ) *P ) ;
		*P = NULL ;
	}
}

static void free_all()
{
	free_and_NULL ( ( const void ** ) &devicename );
	free_and_NULL ( ( const void ** ) &output_file );
	free_and_NULL ( ( const void ** ) &data_str );
	free_and_NULL ( ( const void ** ) &dev_codepage );
	if  ( conv_tty2dev )
		iconv_close( conv_tty2dev );
	if  ( conv_dev2tty )
		iconv_close( conv_dev2tty );
}

static void print_help ( FILE* out_stream )
{
	fprintf ( out_stream, _ (
"\n\
ttytest loopback (read/write) test of serial device  <tty_dev>\n\
\n\
Usage: ttytest [options] <tty_dev>\n\
Valid options are:\n\
-h, --help           display this help screen\n\
-V, --version        display version information\n\
-v, --verbose        verbose/debug output\n\
-c, --config-file    use alternative config file (/etc/ttytest.conf by default)\n\
-o, --output-to      redirect output to file ...\n\
-d, --data-string    test string (\"test test test test ...\" used as default) \n\
-e, --enum-strings   enumerate data-string before write it\n\
-u, --no-offset      disable autoinsert periodical offset before test string\n\
-p, --pause          pause before start of test ( no pause by default )\n\
-a, --adaptive-tx    use adaptive transmitter ( support transmit buffer in maximal empty status if -q and -w options has no effect)\n\
-w, --wait-tx        wait complette data transfer\n\
-q, --queue-size     max number of characters in the output queue\n\
-s, --send-interval  time betwheen two sending of test strings ( unterminated stream if not set )\n\
-m, --mode           start in mode loopback,read,stub,terminal,write\n\
-x, --hex-dump       print bytes in hexadecimal ( 0xXX ) notation\n\
-B, --tty-baud       tty baud rate\n\
-W, --tty-databits   tty data bits ( 1..8 )\n\
-S, --tty-stopb      tty stop bits ( 1, 1.5, 2 )\n\
-P, --tty-parity     tty parity check (no,odd,even,mark,space)\n\
-R, --tty-crtscts    tty use RTS/CTS  flow control\n\
-C, --tty-codepage   tty remote codepage (MTK-2 included)\n\
-N, --tty-nonprint_show  tty show nonprintable chars (in MTK encoding)\n\
-U, --mtk-rprefix    tty registry chars in MTK encoding must be prefixed\n\
-X, --tty-xonxoff    tty use Xon/Xoff flow control\n\
-D, --tty-dsr        tty is DSR sensibility\n\n\
Report bugs to \"Sergey Shilov \" <shilov_serg@users.sourceforge.net>\n\n"  ) );
}


static void help()
{
	print_help ( stdout );
	free_all();
	exit ( EXIT_SUCCESS );
}

void usage ()
{
	print_help ( stderr );
	free_all();
	exit ( EX_CONFIG );
}

//
//  Close files and free memory
//  on abnormal program termination
//
void terminate ( int sysexit_code )
{
	if ( verbose )
		printf ( _ ( "Do terminate\n" ) );

	if ( ( tx_tid != ( pthread_t ) -1 ) && ( pthread_cancel ( tx_tid ) == 0 ) )
		pthread_join ( tx_tid, NULL );

	if ( ( rx_tid != ( pthread_t ) -1 ) && ( pthread_cancel ( rx_tid ) == 0 ) )
		pthread_join ( rx_tid, NULL );


	if ( device_w != NULL )
	{
		if ( verbose )
			printf ( _ ( "Close %s write stream " ),  devicename );
#ifdef	C_LINUX

		__fpurge ( device_w );
#endif

//		fclose ( device_w );
		if ( close ( fileno( device_w ) ) == -1 ) 
			fprintf ( stderr,  _ ( "Close of %s write stream error:%s.\n" ),  devicename, strerror(errno) );

		if ( verbose )
			printf (  " OK!\n" );
	}
	if ( device_r != NULL )
	{
		if ( verbose )
			printf ( _ ( "Close %s read stream " ),  devicename );
#ifdef	C_LINUX

		__fpurge ( device_r );
#endif

//		fclose ( device_r );
		if ( close ( fileno( device_r ) ) == -1 ) 
			fprintf ( stderr,  _ ( "Close of %s read stream error:%s.\n" ),  devicename, strerror(errno) );

		if ( verbose )
			printf (  " OK!\n" );
	}
	free_all();
	exit ( sysexit_code );
}

//
//  Handle ^C
//
static void signal_handler ( int sign )
{
	switch ( sign )
	{
		case SIGTERM :
			terminate ( EXIT_SUCCESS );
		case SIGINT :
			STOP=1;
	}
}

//
//  Attach handler procedure to signals
//
void set_sighandlers ()
{
	struct sigaction sa;

	sa.sa_handler = signal_handler;
	sigemptyset ( &sa.sa_mask );
	sa.sa_flags = 0;
	sigaction ( SIGINT,&sa,NULL );
	sigaction ( SIGTERM,&sa,NULL );
	sigaction ( SIGHUP,&sa,NULL );
};

//
//  Standard option parsing
//
int prepare_opts ( int argc, char *argv[] )
{
	int ret;

	for ( ;; )
	{
		int opt_index = 0;
		const char *args = "hVvc:d:eup:awq:s:m:xlo:B:W:S:P:C:NURXD";
		static struct option long_options[] =
		    {
			    {
				    "help", 0, NULL, 'h'
			    },
			    {
			        "version", 0, NULL, 'V'
			    },
			    {
			        "verbose", 0, NULL, 'v'
			    },
			    {
			        "config-file", 1, NULL, 'c'
			    },
			    {
			        "output-to", 1, NULL, 'o'
			    },
			    {
			        "data-string", 1, NULL, 'd'
			    },
			    {
			        "enum-strings", 1, NULL, 'e'
			    },
			    {
			        "no-offset", 0, NULL, 'u'
			    },
			    {
			        "pause", 1, NULL, 'p'
			    },
			    {
			        "queue-size", 1, NULL, 'q'
			    },
			    {
			        "adaptive-tx", 0, NULL, 'a'
			    },
			    {
			        "wait-tx", 0, NULL, 'w'
			    },
			    {
			        "send-interval", 1, NULL, 's'
			    },
			    {
			        "mode", 1, NULL, 'm'
			    },
			    {
			        "hex-dump", 0, NULL, 'x'
			    },
			    {
			        "tty-baud", 1, NULL, 'B'
			    },
			    {
			        "tty-databits", 1, NULL, 'W'
			    },
			    {
			        "tty-stopb", 1, NULL, 'S'
			    },
			    {
			        "tty-parity", 1, NULL, 'P'
			    },
			    {
			        "tty-codepage", 1, NULL, 'C'
			    },
			    {
			        "tty-nonprint_show", 0, NULL, 'N'
			    },
			    {
			        "tty-mtk-rprefix", 0, NULL, 'U'
			    },
			    {
			        "tty-crtscts", 0, NULL, 'R'
			    },
			    {
			        "tty-xonxoff", 0, NULL, 'X'
			    },
			    {
			        "tty-dsr", 0, NULL, 'D'
			    },
			    {
			        NULL, 0, NULL, '\0'
			    }
		    };

		ret = getopt_long ( argc, argv, args, long_options, &opt_index );
		int f_ret = 0 ;

		if ( ret == -1 )
			break;
		else if ( ret == 0 )
			ret = long_options[opt_index].val;

		switch ( ret )
		{
			case 'h':
				help();
				return 0;
			case 'V':
				puts ( VERSION );
				return 0;
			case 'v':
				verbose = 1;
				break;
			case 'c':
				config_file = strdup ( optarg );
				fprintf ( stderr,  _ ( "Option -c is not implemented.\n" ) );
				terminate ( EX_CONFIG );
				break;
			case 'o':
				output_file = strdup ( optarg );
				fprintf ( stderr,  _ ( "Option -o is not implemented.\n" ) );
				terminate ( EX_CONFIG );
				break;
			case 'd':
				data_str = strdup ( optarg );
				break;
			case 'e':
				enum_strings = 1;
				break;
			case 'u':
				no_offset = 1;
				break;
			case 'p':
				f_ret = sscanf ( optarg,"%i",&start_pause );
				if ( f_ret <= 0 )
					usage ();
				break;
			case 'q':
				f_ret = sscanf ( optarg,"%i",&queue_size );
				if ( f_ret <= 0 )
					usage ();
				break;
			case 'w':
				wait_tx = 1;
				break;
			case 'a':
				adaptive_tx = 1;
				break;
			case 's':
				f_ret = sscanf ( optarg,"%f",&send_interval );
				if ( f_ret <= 0 )
					usage ();
				break;
			case 'm':
				switch ( optarg[0] )
				{
					case 'L':
					case 'l':
						mode = MODE_LOOPBACK;
						break;
					case 'R':
					case 'r':
						mode = MODE_READ;
						break;
					case 'S':
					case 's':
						mode = MODE_STUB;
						break;
					case 'T':
					case 't':
						mode = MODE_TERMINAL;
						break;
					case 'W':
					case 'w':
						mode = MODE_WRITE;
						break;
					default:
						fprintf ( stderr,  _ ( "Invalid mode value ( \"%s\" )\n" ), optarg );
						usage ();
				}
				break;
			case 'x':
				hex_print = 1;
				break;
			case 'B':
				f_ret = sscanf ( optarg,"%li",&baud_rate );
				if ( f_ret <= 0 )
					usage ();
				break;
			case 'W':
				f_ret = sscanf ( optarg,"%li",&DATABITS );
				if ( ( f_ret <= 0 ) || ( DATABITS < 4 ) || ( DATABITS > 8 ) )
					usage ();
				break;
			case 'S':
				f_ret = sscanf ( optarg,"%f",&VSTOPBITS );
				if ( ( f_ret <= 0 ) || ! ( ( VSTOPBITS == 1 ) || ( VSTOPBITS == 1.5 ) || ( VSTOPBITS == 2 ) ) )
					usage ();
				break;
			case 'P':
				{
					switch ( optarg[0] )
					{
						case 'N':
						case 'n':
							PARITYON = 0;
							break;
						case 'O':
						case 'o':
							PARITYON = PARENB;
							PARITY = PARODD;
							break;
						case 'E':
						case 'e':
							PARITYON = PARENB;
							PARITY = 0;
							break;
						case 'M':
						case 'm':
							fprintf ( stderr, _ ( "Parity Mark not supported\n" ) );
							break;
						case 'S':
						case 's':
							fprintf ( stderr, _ ( "Parity Space not supported\n" ) );
							break;
						default:
							fprintf ( stderr, _ ( "Invalid Parity value\n" ) );
							usage ();
					}
				}
			case 'C':
				dev_codepage = strdup ( optarg );
				if (( dev_codepage == NULL ) || ( strlen ( dev_codepage ) < 2 ))
				{
					fprintf ( stderr, _ ( "Invalid  codepage value.\n" ) );
					usage ();
				}
				// Check MTK2 encoding hack.
				if ( ( strstr (dev_codepage, "MTK") != NULL ) || ( strstr (dev_codepage, "mtk") != NULL ) ) 
				{
					MTK2_IO = 1;
					free (dev_codepage);
					dev_codepage = strdup ( "CP866" );
				}
				break;
			case 'N':
				show_nonprint = 1;
				break;
			case 'U':
				use_mtk_reg_prefix = 1;
				break;
			case 'R':
				use_RTSCTS = 1;
				break;
			case 'X':
				use_XonXoff = 1;
				break;
			case 'D':
				use_DSR = 1;
				break;
			default:
				{
					usage ();
				}
		}
	}
	if ( optind < argc )
		devicename = strdup ( argv[optind++] );
	else
	{
		printf ( _ ( "Required device name is not set!\n" ) );
		usage ();
	}
	return 0;
}
//
// RX thread loop procedure
//
extern void *rx_thread_proc ( void *rx_args )
{
	const unsigned int max_buf_size = 16;
	char rbuf[max_buf_size];
	char cbuf[max_buf_size];
	char * p_rbuf = NULL;
	char * p_cbuf = NULL;
	size_t s_rbuf = 0;
	size_t s_cbuf = 0;

	cmtk_conv_data_t rconv_data = def_cmtk_conv_data;
	rconv_data.flag = show_nonprint ;

	int retval;
	int i;

	struct thread_args *args = ( struct thread_args * ) rx_args;
	int qsize;

	fd_set rfds;
	struct timeval tv;
	int dev_fd = fileno ( args->dev_r );


	// Enable "hard" thread cancel
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	while ( ( STOP_RX == 0 ) && ( ! feof ( args->dev_r ) ) )
	{
		FD_ZERO ( &rfds );
		FD_SET ( dev_fd, &rfds );
		// Test - is device readable
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		retval = select ( dev_fd + 1, &rfds, NULL, NULL, &tv );
		if ( retval == -1 )
		{
			if ( errno == EINTR ) // interrupted system call
				fprintf ( stderr, _ ( "\n%s: RX interrupt\n" ), devicename );
			else
				fprintf ( stderr, _ ( "%s: rx read error: %s\n" ), devicename, strerror ( errno ) );
			break;
		}
		else 	if ( retval == 0 ) 
		{
			if ( verbose ) //Permanently test and print size of queue (in verbose  mode)
				fprintf ( stdout, _ ( "%s: rx select ( read ) timeout.\n" ), devicename );
			continue;
		}
		//****************************************************
		//
		//	For test of a TIOCINQ ioctl only
		//
		if ( verbose )
		{
			if (  ioctl ( ( int ) fileno ( args->dev_r ), TIOCINQ, &qsize ) < 0 )
				fprintf ( stderr, _ ( "%s:  TIOCINQ ioctl error: %s\n" ), devicename, strerror ( errno ) );
			else
				fprintf ( stdout, "%s: TIOCINQ=%i\n", devicename, qsize);
		}
		//****************************************************
		p_rbuf = (char *) &rbuf[0];
		s_rbuf = 0;
		p_cbuf = (char *) &cbuf[0];
		s_cbuf = max_buf_size;
		memset(p_rbuf, 0x00, max_buf_size);
		memset(p_cbuf, 0x00, max_buf_size);

		s_rbuf = fread ( p_rbuf, 1, 1, args->dev_r );
		//s_rbuf = read ( dev_fd, p_rbuf, 1 );
		if ( s_rbuf > 0 )
		{
			if ( ( args->mode == MODE_STUB ) && ( args->dev_w != NULL ) )
			{
				if ( fputc ( ( int ) rbuf[0], args->dev_w ) == EOF )
				{
					fprintf ( stderr, _ ( "%s: write data error: %s\n" ), devicename, strerror ( errno ) );
					terminate ( EX_SOFTWARE );
				}
				fflush ( args->dev_w );
			}
			if ( dev_codepage != NULL )
			{
				if ( MTK2_IO )
				{
					p_rbuf[1] = 0xff;
					s_rbuf = cmtk_buf_MTK2toCP866( &rconv_data, ( unsigned char *) p_rbuf, s_rbuf, p_cbuf, max_buf_size);
					memcpy ( p_rbuf, p_cbuf, max_buf_size );
				}
				retval = iconv ( conv_dev2tty,  &p_rbuf, &s_rbuf, &p_cbuf, &s_cbuf);
			}
			else
				memcpy ( p_cbuf, p_rbuf, s_rbuf  );
		}
		else
		{
			if ( ferror ( args->dev_r ) && ( errno != EINTR ) )
			{
				fprintf ( stderr, _ ( "%s: rx read error: %s\n" ), devicename, strerror ( errno ) );
				break;
			}
			else
				continue;
		}
		s_cbuf = strlen( &cbuf[0] );
		for( i = 0 ; i < s_cbuf ; i++ )
		{
			if ( hex_print == 1 )
				retval = ( fprintf ( args->out, "0x%2.2x\n", ( unsigned char ) cbuf[i] ) > 0 ) ? 1 : EOF;
			else
				retval = ( ( fputc ( ( int ) cbuf[i], args->out ) == EOF ) && ( ferror ( args->out ) ) )  ?  EOF : 1;
			if ( retval == EOF )
			{
				fprintf ( stderr, _ ( "%s: rx output error: %s\n" ), devicename, strerror ( errno ) );
				break;
			}
			fflush ( args->out );
		}
		if ( retval == EOF )
			break;
	}
	if ( verbose )
		fprintf ( stdout, _ ( "\n%s: stop RX OK!\n" ), devicename );
	pthread_exit ( 0 );
}
//
// TX thread loop procedure
//
extern void *tx_thread_proc ( void *tx_args )
{

	const int max_offset = 9;
	const int max_data_str_len = 256;

	int dev_fd = -1;

	struct thread_args *args;
	args = ( struct thread_args * ) tx_args;

	char data_str_buf[max_data_str_len];
	char cdata_str_buf[max_data_str_len];

	memset ( &data_str_buf, 0, max_data_str_len);
	memset ( &cdata_str_buf, 0, max_data_str_len);

	char offset_buf[max_offset+1];

	cmtk_conv_data_t rconv_data , wconv_data;
	rconv_data = wconv_data = def_cmtk_conv_data;

	if ( use_mtk_reg_prefix == 0 )
	{
		wconv_data.regs[0].prefix_char = wconv_data.regs[1].prefix_char = wconv_data.regs[2].prefix_char = 0x00;
		wconv_data.regs[0].reg_chars[1] = wconv_data.regs[1].reg_chars[1] = wconv_data.regs[2].reg_chars[1] = 0x00;
	}
	char * out_str = NULL;    // Warriors printed data

	// Internal variables
	int n = 0;
	int len = 0;
	int offset = max_offset;
	long num = 1;
	int qsize = 0;
	int dsize = 0;
	long s_usec = 0;

	int data_str_buf_used = 0;

	int retval = -1;
	fd_set wfds;
	struct timeval tv;

	time_t start ;
	time ( &start );
	double seconds_left = 0;
	long tx_real = 0;
	long tx_left = 0;

	long tx_do_total = 0;
	long tx_total = 0;
	long tx_delta = 0;

	struct serial_icounter_struct counters_old;
	struct serial_icounter_struct counters_new;
	memset ( &counters_old, 0, sizeof ( counters_old ) );
	memset ( &counters_new, 0, sizeof ( counters_new ) );
	//----------------------------------------------------------------

	memset ( &offset_buf[0], 0x20, max_offset );
	offset_buf[max_offset] = '\0';

	dev_fd = fileno ( args->dev_w );

	// Enable "hard" thread cancel
	if ( ! args->flags && FLAG_IN_MAIN_THREAD )
		pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	while ( ( STOP_TX == 0 ) && ( ! feof ( args->dev_w ) ) )
	{
		retval = -1;
		FD_ZERO ( &wfds );
		FD_SET ( dev_fd, &wfds );

		memset ( &data_str_buf[0], 0, max_data_str_len );
		memset ( &cdata_str_buf[0], 0, max_data_str_len );
		data_str_buf_used = 0;

		// Create delay between two test string ( --send-interval option )
		if ( send_interval > 0 )
			usleep ( send_interval * 1000000 );

		// Test - is device writable
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		retval = select ( dev_fd + 1, NULL, &wfds, NULL, &tv );

		if ( retval == -1 )
		{
			if ( errno == EINTR ) // interrupted system call
				fprintf ( stderr, _ ( "\n%s: TX interrupt\n" ), devicename );
			else
				fprintf ( stderr, _ ( "%s: tx write error: %s\n" ), devicename, strerror ( errno ) );

			if ( ! args->flags && FLAG_IN_MAIN_THREAD )
				pthread_mutex_lock ( &STOP_TX_mutex );
			STOP_TX = 1;
			if ( ! args->flags && FLAG_IN_MAIN_THREAD )
				pthread_mutex_unlock ( &STOP_TX_mutex );
			continue;
		}
		else 	if ( ( retval == 0 ) && ( STOP_TX == 0 ) )
		{
			if ( verbose ) //Permanently test and print size of queue (in verbose  mode)
			{
				fprintf ( stdout, _ ( "%s: tx select ( write ) timeout.\n" ), devicename );
			}
			continue; // Read timeout ( wfds==NULL  - device busy?)
		}

		if ( ( args->mode == MODE_TERMINAL ) && ( args->in != NULL ) )
		{
			if (  fgets ( (char *) &data_str_buf, max_data_str_len, args->in ) == NULL )
			{
				if ( ( ferror (args->in) ) && ( errno != EINTR ) )
				{
					fprintf ( stderr, _ ( "%s: read input error: %s\n" ), devicename, strerror ( errno ) );
					terminate ( EX_SOFTWARE );
				}
				else if  ( feof (args->in) )
				{
					if ( ! args->flags && FLAG_IN_MAIN_THREAD )
						pthread_mutex_lock ( &STOP_TX_mutex );
					STOP_TX = 1;
					if ( ! args->flags && FLAG_IN_MAIN_THREAD )
						pthread_mutex_unlock ( &STOP_TX_mutex );
					continue;
				}
				else
					continue;
			}

		}
		else
		{
			//----------------------------------------------------------
			// Build a test string
			// - insert enumerator
			// - move to dynamic offset
			out_str = &data_str_buf[0];
			if ( enum_strings )
			{
				if ( num == 999999 )
					num = 1;
				data_str_buf_used = snprintf ( data_str_buf, max_data_str_len - 1, "%6li ", num );
				out_str = &data_str_buf[data_str_buf_used];
				num++;
			}
			if ( no_offset )
			{
				data_str_buf_used = snprintf ( out_str, max_data_str_len - 1 - data_str_buf_used, "%s\n", data_str );
			}
			else
			{
				snprintf ( out_str, max_data_str_len - 1 - data_str_buf_used, "%s%s\n", ( char * ) &offset_buf[offset], data_str );
				if ( offset == 0 )
					offset = max_offset;
				else
					offset--;
			}
		}
		out_str = &data_str_buf[0];
		//----------------------------------------------------------
		if ( dev_codepage != NULL )
		{
			const char * p_in = (char *) &cdata_str_buf;
			const char * p_out = (char *) &data_str_buf;
			size_t inl = 0 , outl = 0;
			memcpy( ( void * ) p_in,  ( void * ) p_out, max_data_str_len);
			inl = strlen( (char * ) &cdata_str_buf );
			outl = max_data_str_len;
			retval = iconv ( conv_tty2dev, (char ** ) &p_in, &inl, (char ** )  &p_out, &outl );
			len = max_data_str_len - outl;
			data_str_buf[len] = '\0';
			if ( MTK2_IO == 1 )
			{
				memcpy( ( void * ) &cdata_str_buf,  ( void * ) &data_str_buf, max_data_str_len);
				inl = len;
				len = cmtk_buf_CP866toMTK2( &wconv_data,
			              (unsigned char *) &cdata_str_buf[0], inl,
			              (unsigned char *) &data_str_buf[0], max_data_str_len);
			}
		}
		else
			len = strlen ( out_str );

		out_str = &data_str_buf[0];
		// Write output test string to serial device loop
		do
		{
			// -----------------------------------------------------------------------------------------
			// This part of code illystrate a problem
			// with big chips-buffers on low...low speed
			// terminals (B50..B1200 bauds)
			//
			// TIOCOUTQ ioctl command used to lock input
			// while all data PHYSICAL not trancferred rfom buffers
			// The size of queue can sets by --queue-size option
			//
			//
			// Test output queue
			retval = ioctl ( dev_fd, TIOCOUTQ, &qsize );
			if ( retval < 0 )
			{
				fprintf ( stderr, _ ( "%s: ioctl(%i,  TIOCOUTQ, ... ) error: %s\n" ), devicename, dev_fd, strerror ( errno ) );
				if ( ! args->flags && FLAG_IN_MAIN_THREAD )
					pthread_mutex_lock ( &STOP_TX_mutex );
				STOP_TX = 1;
				if ( ! args->flags && FLAG_IN_MAIN_THREAD )
					pthread_mutex_unlock ( &STOP_TX_mutex );
				continue;
			}
			//
			//
			dsize = len;

			if ( wait_tx )
			{
				if ( ( tcdrain ( dev_fd ) != 0 ) && ( errno != EINTR ))
					fprintf ( stderr,_ ( "%s: wait of data transfer error code %i : %s\n" ),
					          devicename, retval, strerror ( errno ) );
				fflush ( args->dev_w );
			}
			else	if ( adaptive_tx )
			{
				// In adaptive TX mode the transmitter wait in
				// loop - while all stored  data into buffers
				// is transferred to out.
				while ( ( tx_do_total > tx_real ) && ( STOP_TX == 0 ) )
				{
					// Get serial port counters into  serial_icounter_struct structure
					memset ( &counters_new, 0, sizeof ( counters_new ) );
					retval = ioctl ( dev_fd, TIOCGICOUNT, &counters_new );
					if ( retval < 0 )
					{
						fprintf ( stderr, _ ( "%s: ioctl(%i, TIOCGICOUNT, ... ) error: %s\n" ),
						          devicename, dev_fd, strerror ( errno ) );
						if ( ! args->flags && FLAG_IN_MAIN_THREAD )
							pthread_mutex_lock ( &STOP_TX_mutex );
						STOP_TX = 1;
						if ( ! args->flags && FLAG_IN_MAIN_THREAD )
							pthread_mutex_unlock ( &STOP_TX_mutex );
						continue;
					}
					// Calculate tx delta as .tx counter delta
					if ( counters_old.tx > 0 ) //
						tx_delta = counters_new.tx - counters_old.tx;
					else
						tx_delta = 0; // if first iteration - reset counter
					//
					if ( tx_delta > 0 )
						tx_total = tx_total + tx_delta;
					// Remember last state of counters
					counters_old = counters_new;

					// Calculate ~ transferred bytes count ( tx time * speed in Bytes )
					time_t now ;
					time ( &now ) ;
					seconds_left = difftime ( now, start );
					tx_real = ( seconds_left * baud_rate ) /8 ;
					// Calculate ~ untransferred data length
					tx_left = tx_do_total > tx_real ? tx_do_total - tx_real : 0 ;
					if ( verbose )
						fprintf ( stdout,"%s: tx_delta=%li ( tx_do_total=%li, tx_total=%li, tx_real~=%li, tx_left=~%li)\n",
						          devicename, tx_delta, tx_do_total, tx_total, tx_real, tx_left );
					if ( tx_left > 0 )
					{
						struct timeval tp0, tp1, tpd =
						    {
							    0, 0
						    };
						if ( verbose )
						{
							gettimeofday ( &tp0, NULL );
							double s_sec = ( double ) s_usec / 1000000 ;
							fprintf ( stdout, _ ( "%s: sleep %.6f S \n" ),   devicename, s_sec );
						}
						// Calculate delay witch needed to send untransferred data really
						s_usec = ( ( 1000000 / baud_rate ) * 8 )  * tx_left ;
						//Wait transfer
						tv.tv_sec=s_usec/1000000;
						tv.tv_usec= s_usec % 1000000;
						retval = select ( 0, NULL, NULL, NULL, &tv );
						if ( verbose )
						{
							gettimeofday ( &tp1, NULL );
							tpd.tv_sec = tp1.tv_sec - tp0.tv_sec;
							if ( tp1.tv_usec < tp0.tv_usec )
							{
								tpd.tv_usec = tp1.tv_usec + 1000000 - tp0.tv_usec;
								tpd.tv_sec--;
							}
							else
								tpd.tv_usec = tp1.tv_usec - tp0.tv_usec;
							fprintf ( stdout, _ ( "\n%s: sleep/wait of %i.%i S done.\n" ),   devicename,
							          ( int ) tpd.tv_sec, ( int ) tpd.tv_usec );
						}
					}
				}
			}
			// If need the monitoring of size of output queue
			else if ( queue_size >= 0 )
			{
				if ( verbose )
					fprintf ( stdout, _ ( "%s: TIOCOUTQ=%i. (Rreally writted %li Bytes )\n" ),  devicename, qsize, tx_do_total );
				// Check size of output queue
				if ( qsize > queue_size )
				{
					// If real datasize in buffer up to --queue-size limit then
					// calculate time of real transfer qsize-bytes from queue
					s_usec = ( 1000000 / baud_rate ) * 8 /* DATABITS ? */ * qsize ;
					if ( verbose )
					{
						double s_sec = ( double ) s_usec / 1000000 ;
						fprintf ( stdout, _ ( "%s: wait %0.2fS for queue transfer.\n" ),  devicename,  s_sec );
					}
					// and block/delay transfer to...
					// Wait to transfer
					tv.tv_sec=0;
					tv.tv_usec=s_usec;
					retval = select ( 0, NULL, NULL, NULL, &tv );
					continue;
				}
				else
				{
					// Set size of writted data to 1 or queue_size - qsize
					dsize = qsize == queue_size ? 1 : queue_size - qsize ;
					if ( dsize > len )
						dsize = len;
				}
			}
			// -----------------------------------------------------------------------------------------
			n = fwrite ( out_str, 1, dsize, args->dev_w );
			if ( n > 0 )
			{
				if ( verbose )
				{
					tx_do_total = tx_do_total + n;
					fprintf ( stdout, _ ( "%s: write %i Bytes ( %i writted)\n" ),  devicename, dsize, n );
				}
				if ( args->flags && FLAG_PRINT_OUTPUT )
				{
					while ( n > 0 )
					{
						// reconvert by char
						if ( dev_codepage != NULL )
						{
							if ( MTK2_IO)
							{
								char rbuf[16];
								char cbuf[16];
								char * p_rbuf = &rbuf[0];
								char * p_cbuf = &cbuf[0];
								size_t s_rbuf = 16;
								size_t s_cbuf = 0;
								cbuf[1] = 0xff;
								s_cbuf = cmtk_buf_MTK2toCP866( &wconv_data, (unsigned char *) out_str, 1, p_cbuf, 16);
								if ( s_cbuf > 0 )
								{
									memset ( p_rbuf, 0, 16 );
									retval = iconv ( conv_dev2tty,  &p_cbuf, &s_cbuf, &p_rbuf, &s_rbuf);
									p_rbuf = &rbuf[0];
									while ( *p_rbuf != 0x00 )
									{
//										fprintf ( args->out, "%1s  (0x%2.2x)\n", p_rbuf, *p_rbuf );
										fputc ( *p_rbuf, args->out );
										p_rbuf++;
									}
								}
							}
							else
								fputc ( cdata_str_buf[ out_str - &data_str_buf[0] ], args->out );
						}
						else
							fputc ( ( char ) *out_str, args->out );
						len --;
						out_str ++;
						n --;
					}
				}
				else
				{
					len -= n;
					out_str += n;
				}
			}
			else if ( feof ( args->dev_w ) )
			{
				fprintf ( stderr, _ ( "%s: connection refused (EOF)\n" ), devicename );
				if ( ! args->flags && FLAG_IN_MAIN_THREAD )
					pthread_mutex_lock ( &STOP_TX_mutex );
				STOP_TX = 1;
				if ( ! args->flags && FLAG_IN_MAIN_THREAD )
					pthread_mutex_unlock ( &STOP_TX_mutex );
			}
			else if ( ferror ( args->dev_w ) )
			{
				fprintf ( stderr, _ ( "%s: write error: %s\n" ), devicename, strerror ( errno ) );
				if ( ! args->flags && FLAG_IN_MAIN_THREAD )
					pthread_mutex_lock ( &STOP_TX_mutex );
				STOP_TX = 1;
				if ( ! args->flags && FLAG_IN_MAIN_THREAD )
					pthread_mutex_unlock ( &STOP_TX_mutex );
			}
			else
			{
				fprintf ( stderr, _ ( "%s: write error or not error?. No data writte and no error returned.\n" ), devicename );
				continue;
			}
		}
		while ( ( len > 0 ) && ( STOP_TX == 0 ) );
	}
#ifdef	C_LINUX

	__fpurge ( args->dev_w );
#endif
	if ( verbose )
		fprintf ( stdout, _ ( "\n%s: stop TX OK!\n" ), devicename );
	pthread_exit ( 0 );
}


int start_thread ( int tread_type, struct thread_args *arg )
{
	char * stread_name_rx = "rx";
	char * stread_name_tx = "tx";

	int ret = -1;
	char *stread_name = NULL;
	void * thread_proc = NULL;

	pthread_attr_t * p_attr = NULL;
	pthread_t * p_tid = NULL;

	//rx_arg, tx_arg;
	switch ( tread_type )
	{
		case THREAD_RX:
			{
				stread_name = stread_name_rx;
				p_attr = &rx_tattr;
				p_tid = &rx_tid;
				thread_proc = rx_thread_proc;
				break;
			}
		case THREAD_TX:
			{
				stread_name = stread_name_tx;
				p_attr = &tx_tattr;
				p_tid = &tx_tid;
				thread_proc = tx_thread_proc;
				break;
			}
		default:
			{
				fprintf ( stderr, _ ( "%s: thread initialization failed. invalid thread type\n" ), devicename);
				terminate ( EX_SOFTWARE );
			}
	}

	ret = pthread_attr_init ( p_attr );
	if ( ret != 0 )
	{
		fprintf ( stderr, _ ( "%s:init of %s thread error: %s\n" ), devicename, stread_name, strerror ( errno ) );
		terminate ( EX_OSERR );
	}
	ret = pthread_create ( p_tid, p_attr, thread_proc, arg );
	if ( ret != 0 )
	{
		fprintf ( stderr, _ ( "%s:init of %s thread error: %s\n" ), devicename, stread_name, strerror ( errno ) );
		terminate ( EX_SOFTWARE );
	}
	return ret;
}


int stop_thread ( int tread_type )
{
	int ret = -1;
	char *stread_name = NULL;

	switch ( tread_type )
	{
		case THREAD_RX:
			{
				stread_name = "rx";
				pthread_mutex_lock ( &STOP_RX_mutex );
				STOP_RX = 1;
				pthread_mutex_unlock ( &STOP_RX_mutex );
				ret = pthread_join ( rx_tid, NULL );
				break;
			}
		case THREAD_TX:
			{
				stread_name = "tx";
				pthread_mutex_lock ( &STOP_TX_mutex );
				STOP_TX = 1;
				pthread_mutex_unlock ( &STOP_TX_mutex );
				ret = pthread_join ( tx_tid, NULL );
				break;
			}
		default:
			{
				fprintf ( stderr, _ ( "%s: thread initialization failed. invalid thread type\n" ), devicename);
				terminate ( EX_SOFTWARE );
			}
	}
	if ( ret != 0 )
		fprintf ( stderr, _ ( "%s: join of %s thread error: %s\n" ), devicename, stread_name, strerror ( errno ) );
	return ret;
}

int main ( int argc, char *argv[] )
{
	struct termios old_tty_params, new_tty_params;
	struct termios old_output_params, new_output_params;
	struct termios old_keyboard_params, new_keyboard_params;
	struct thread_args rx_arg, tx_arg;

#ifdef	C_LINUX

	setlocale ( LC_ALL, "" );
	bindtextdomain ( argv[0], "/usr/share/locale" );
	textdomain ( argv[0] );
#endif

	if ( verbose )
		fprintf ( stdout, _ ( "Init internal variables.\n" ) );
	// Clear buffers from trush
	memset ( &old_tty_params, 0, sizeof ( old_tty_params ) );
	memset ( &new_tty_params, 0, sizeof ( new_tty_params ) );
	memset ( &old_output_params, 0, sizeof ( old_output_params ) );
	memset ( &new_output_params, 0, sizeof ( new_output_params ) );
	memset ( &old_keyboard_params, 0, sizeof ( old_keyboard_params ) );
	memset ( &new_keyboard_params, 0, sizeof ( new_keyboard_params ) );
	memset ( &rx_arg, 0, sizeof ( rx_arg ) );
	memset ( &tx_arg, 0, sizeof ( tx_arg ) );

	// If test string NOT set from command line, use default value
	if ( data_str == NULL )
		data_str = strdup ( _ ( DEF_DATA_STRING ) );

	// Pre set output as stdout (before config read)
	output = stdout;

	prepare_opts ( argc, argv );

	if ( verbose )
		fprintf ( stdout, _ ( "Setup conversion.\n" ) );
	// Init conversion (if needed )
	if ( dev_codepage != NULL )
	{
		// Transform value dev_codepage no upper case (tty_codepage used as "temp" )
		tty_codepage = dev_codepage;
		while ( *tty_codepage != '\0' )
		{
			*tty_codepage = toupper (*tty_codepage);
			tty_codepage ++ ;
		}
		// Get codepage of current locale frpm LANG environmant variable
		tty_codepage = getenv ( "LANG" );
		if ( tty_codepage != NULL )
		{
			
			char * s_locale = strrchr( tty_codepage,  '.' ) ;
				
			if  ( s_locale != NULL )
			{
				s_locale++; // skip '.' in variable value
				tty_codepage = strdup( s_locale ) ; 
			}
			else
			{
				FILE * cf = fopen ( "/etc/sysconfig/consolefont", "r" );
				if ( cf != NULL )
				{
					char rbuf[128];
					tty_codepage = NULL;
					while ( ! feof ( cf ) )
					{
						memset ( &rbuf, 0, 128  );
						if ( fgets ( ( char * ) &rbuf, 128, cf ) != NULL )
						{
							if ( strstr( ( char * ) &rbuf, "SYSFONTACM" ) != NULL )	//SYSFONTACM=.....
							{
								s_locale = strrchr( ( char * ) &rbuf,  '=' );
								if (s_locale == NULL )
									continue;
								int i=0, j=0 , len = strlen(( char * ) &rbuf);
								for ( i = 0, j = ( s_locale - ( char * ) &rbuf) ; j < len ; j ++ )
								{
									s_locale++;
									if ( ! isspace( *s_locale ) )
										rbuf[i++] = toupper ( *s_locale ) ;
								}
								rbuf[i++] = '\0';
								tty_codepage = strdup(( char * ) &rbuf) ;
								break;
							}
						}
					}
					if ( tty_codepage ==  NULL)
					{
						fprintf ( stderr, _ ( "Cannot get SYSFONTACM value.\n" ) );
						terminate ( EX_CONFIG );
					}
					fclose(cf);
				}
				else
				{
					fprintf ( stderr, _ ( "Cannot get current locale from $LANG (%s) variable.\n" ), tty_codepage );
					terminate ( EX_CONFIG );
				}
			}	
			if ( verbose )
				fprintf ( stderr, _ ( "%s: Init covertors tty=%s <> dev=%s \n" ), devicename, tty_codepage, dev_codepage);
			// Init convertors
			//
			conv_tty2dev = iconv_open(dev_codepage, tty_codepage);
			if ( conv_tty2dev == ( ( iconv_t )( -1 ) ) )
				fprintf ( stderr, "The  conversion  from %s to %s is not supported by the system.",
				          tty_codepage, dev_codepage);
			conv_dev2tty = iconv_open(tty_codepage, dev_codepage);
			if ( verbose )
				fprintf ( stderr, "Init internal I/O convertor DEV(%s) TTY(%s)\n", dev_codepage,  tty_codepage);
		}
		else
		{
			fprintf ( stderr, _ ( "Cannot determinate system LANGuage. Get $LANG value error: %s\n" ), strerror ( errno ) );
			terminate ( EX_OSERR );
		}
	}

	if ( verbose )
		fprintf ( stderr, _ ( "%s: Stup parameters...\n" ), devicename );

	//  Normalize baud
	switch ( baud_rate )
	{
		case 50:
			BAUD  = B50;
			break;
		case 75:
			BAUD  = B75;
			break;
		case 110:
			BAUD  = B110;
			break;
		case 134:
			BAUD  = B134;
			break;
		case 150:
			BAUD  = B150;
			break;
		case 200:
			BAUD  = B200;
			break;
		case 300:
			BAUD  = B300;
			break;
		case 600:
			BAUD  = B600;
			break;
		case 1200:
			BAUD  = B1200;
			break;
		case 1800:
			BAUD  = B1800;
			break;
		case 2400:
			BAUD  = B2400;
			break;
		case 4800:
			BAUD  = B4800;
			break;
		case 9600:
			BAUD  = B9600;
			break;
		case 19200:
			BAUD  = B19200;
			break;
		case 38400:
			BAUD = B38400;
			break;
		default:
			BAUD  = B9600;
	}

	switch ( DATABITS )
	{
		case 5:
			DATABITS = CS5;
			break;
		case 6:
			DATABITS = CS6;
			break;
		case 7:
			DATABITS = CS7;
			break;
		case 8:
		default:
			DATABITS = CS8;
			break;
	}

	if ( VSTOPBITS == 2 )
		STOPBITS = CSTOPB;
	else
		STOPBITS = 0;

	//Setup signals
	set_sighandlers ();

	// Ignore SIGIO signal (used select() function)
	signal ( SIGIO, SIG_IGN );

	if ( verbose )
		fprintf ( stdout, _ ( "%s: Prepare device...\n" ),  devicename );

	device_w = fopen ( devicename, "w" );
	if ( device_w == NULL )
	{
		fprintf ( stderr, _ ( "Open for writting device %s error: %s\n" ), devicename, strerror ( errno ) );
		return EXIT_FAILURE;
	}

	if ( verbose )
		fprintf ( stdout, _ ( "%s: open write OK!\n" ),  devicename );

	setvbuf(device_w, (char *)NULL, _IONBF, 0);

	device_fd = fileno ( device_w );

	// Save parameters of terminal of serial device
	tcgetattr ( device_fd, &old_tty_params );

	// Set ASYNC I/O
	fcntl ( device_fd, F_SETFL, FASYNC );

	// Setup serial device as terminal
	cfmakeraw ( &new_tty_params );
	new_tty_params.c_cflag = BAUD | CRTSCTS | DATABITS | STOPBITS | PARITYON | PARITY | CREAD | HUPCL | CLOCAL;
	new_tty_params.c_lflag &= ~ICANON;
	new_tty_params.c_lflag &= ~ISIG;
	new_tty_params.c_cc[VMIN]=1;

	if ( tcsetattr ( device_fd, TCSANOW, &new_tty_params ) != 0 )
	{
		fprintf ( stderr, _ ( "%s: setup tty prams error: %s\n" ), devicename, strerror ( errno ) );
		terminate ( EX_SOFTWARE );
	}
	else
		if ( verbose )
			fprintf ( stdout, _ ( "%s: setup as tty OK!\n" ),  devicename );

	// Flush device buffers
	tcflush ( device_fd, TCIOFLUSH );

	device_r = fopen ( devicename, "r" );
	if ( device_r == NULL )
	{
		fprintf ( stderr, _ ( "Open for reading device %s error: %s\n" ), devicename, strerror ( errno ) );
		fclose ( device_w );
		return EXIT_FAILURE;
	}
	
	setvbuf(device_r, (char *)NULL, _IONBF, 0);

	if ( verbose )
		fprintf ( stdout, _ ( "%s: open read OK!\n" ),  devicename );


	// Save output device parameters
	output_fd = fileno ( output );
	tcgetattr ( output_fd, &old_output_params );

	if ( verbose )
		fprintf ( stdout, _ ( "%s: setup MODE\n" ),  devicename );


	rx_arg.mode = tx_arg.mode = mode;

	// Select mode ( --mode option )
	switch ( mode )
	{
		case MODE_READ:
			{
				// Read any data from serial device and write this data backward
				if ( verbose )
					fprintf ( stdout, _ ( "%s: enter in MODE_READ\n" ),  devicename );
				rx_arg.out = output;
				rx_arg.dev_r = device_r;
				rx_arg.dev_w = NULL;
				start_thread ( THREAD_RX, &rx_arg );
				// Main loop
				while ( ( STOP == 0 ) && ( STOP_RX == 0 ))
					sleep ( 1 );
				stop_thread ( THREAD_RX );
				fflush ( output );
				break;
			}
		case MODE_WRITE:
			{
				// Read any data from serial device and write this data backward
				if ( verbose )
					fprintf ( stdout, _ ( "%s: enter in MODE_WRITE\n" ),  devicename );
				tx_arg.out = output;
				tx_arg.dev_r = NULL;
				tx_arg.dev_w = device_w;
				tx_arg.flags |= FLAG_PRINT_OUTPUT;
				start_thread ( THREAD_TX, &tx_arg );
				while ( ( STOP == 0 ) && ( STOP_TX == 0 ))
					sleep ( 1 );
				stop_thread ( THREAD_TX );
				fflush ( output );
				break;
			}
		case MODE_TERMINAL:
			{
				// Simply terminal
				if ( verbose )
					fprintf ( stdout, _ ( "%s: enter in MODE_TERMINAL\n" ),  devicename );
				//------------------------------------------
				// Handle real tty
				input = fopen ( "/dev/tty", "r" );
				if ( input == NULL )
				{
					fprintf ( stderr, _ ( "Unable to open input. Error: %s\n" ), strerror ( errno ) );
					return EXIT_FAILURE;
				}

				output = fopen ( "/dev/tty", "w" );
				if ( output == NULL )
				{
					fprintf ( stderr, _ ( "Unable to open output. Error: %s\n" ), strerror ( errno ) );
					return EXIT_FAILURE;
				}
				//------------------------------------------
				// Start async RX thread
				// redirect readed data to tty
				rx_arg.out = output;
				rx_arg.dev_r = device_r;
				start_thread ( THREAD_RX, &rx_arg );
				//------------------------------------------
				// Scanning keys and write it
				// to serial device in tx thread_args
				// while main program is not terminated
				tx_arg.in = input;
				tx_arg.out = output;
				tx_arg.dev_r = NULL;
				tx_arg.dev_w = device_w;
				tx_arg.flags &= !FLAG_PRINT_OUTPUT;
				send_interval = DEF_SEND_INTERVAL;
				queue_size = -1;
				adaptive_tx = 0;
				wait_tx = 1;
				start_thread ( THREAD_TX, &tx_arg );
				while ( ( STOP == 0 ) && ( STOP_RX == 0 ) /*&& ( STOP_TX == 0 )*/ )
				{
					sleep ( 1 );
				}
				stop_thread ( THREAD_RX );
				stop_thread ( THREAD_TX );
				fclose ( input );
				fclose ( output );
				break;
			}
		case MODE_STUB:
			{
				// Read any data from serial device and write this data backward
				if ( verbose )
					fprintf ( stdout, _ ( "%s: enter in MODE_STUB\n" ),  devicename );
				rx_arg.out = output;
				rx_arg.dev_r = device_r;
				rx_arg.dev_w = device_w;
				start_thread ( THREAD_RX, &rx_arg );
				while ( ( STOP == 0 ) && ( STOP_RX == 0 ) )
					sleep ( 1 );
				stop_thread ( THREAD_RX );
				fflush ( output );
				break;
			}
		case MODE_LOOPBACK:
			{
				// 1. Async write (in TX thread) the test string to serial device ( opened write-only)
				// 2. Async read the test strings (in RX thread) from some device ( opened read-only)
				//    and print results to stdout or -o [file]
				if ( verbose )
					fprintf ( stdout, _ ( "%s: enter in MODE_LOOPBACK\n" ),  devicename );

				// Set output (screen) as  NOT canonical terminal
				new_output_params = old_output_params ;
				new_output_params.c_lflag &= ~ICANON;
				tcsetattr ( output_fd,TCSANOW,&new_output_params );
				// Flush output buffers
				tcflush ( output_fd, TCIOFLUSH );

				// Start RX thread
				rx_arg.out = output;
				rx_arg.dev_r = device_r;
				start_thread ( THREAD_RX, &rx_arg );
				// Start TX thread
				tx_arg.out = output;
				tx_arg.dev_w = device_w;
				start_thread ( THREAD_TX, &tx_arg );

				// Main loop
				while ( ( STOP == 0 ) && ( STOP_RX == 0 ) && ( STOP_TX == 0 ) )
					sleep ( 1 );
				// Stop TX
				stop_thread ( THREAD_TX );
				// Stop RX
				stop_thread ( THREAD_RX );
				tcsetattr ( output_fd,TCSANOW,&old_output_params );
				break;
			}
		default:
			fprintf ( stderr, _ ( "%s: uncnown --mode value\n" ), devicename );
			terminate ( EX_CONFIG );
	}

	// Restore saved tty parameters
	if ( verbose )
		fprintf ( stdout, _ ( "%s: Restore saved parameters.\n" ),  devicename );

	if ( tcsetattr ( device_fd, TCSANOW, &old_tty_params ) != 0 )
	{
		fprintf ( stderr, _ ( "%s: Restore of saved parameters error: %s\n" ), devicename, strerror ( errno ) );
	}
	if ( verbose )
		fprintf ( stdout, _ ( "%s: clean buffers of write file stream\n" ),  devicename );

	//Close device output
#ifdef	C_LINUX

	__fpurge ( device_w );
#endif

	if ( verbose )
		fprintf ( stdout, _ ( "%s: close write file stream\n" ),  devicename );
	if ( fclose ( device_w ) == EOF )
	{
		fprintf ( stderr, _ ( "%s: close device output error: %s\n" ), devicename, strerror ( errno ) );
	}

	//Close device input
	if ( verbose )
		fprintf ( stdout, _ ( "%s: clean buffers of  read file stream\n" ),  devicename );
#ifdef	C_LINUX

	__fpurge ( device_r );
#endif

	if ( verbose )

		fprintf ( stdout, _ ( "%s: close read file stream\n" ),  devicename );
	if ( fclose ( device_r ) == EOF )
	{
		fprintf ( stderr, _ ( "%s: close device input error: %s\n" ), devicename, strerror ( errno ) );
	}

	fprintf ( output, _ ( "\n Exit.\n" ) );
	return EXIT_SUCCESS;
}
